#!/bin/bash

# make a release from a release directory to a staging directory where the
# object to be released (in 'files') is in the same format in the staging
# directory as it was in the release directory

cmd="$0"
cmd_dir="`dirname -- "$cmd"`"
cmd_dir=`(cd "$cmd_dir">/dev/null;pwd)`
cmd_name="`basename -- $cmd`"

release_from="$cmd_dir"

help()
{   echo "syntax: [-v] [-f <rel dir>] [-d|-s] <root stage dir> <boot stage dir>"
    echo "   -v   more verbose output"
    echo "   -d   delete the release held in the stageing directory"
    echo "   -s   select release, copy ops, but do not copy the main release files"
    echo "        otherwise copy the release into the stageing directory"
}

do_help=false
do_delete=false
do_verbose=false
do_no_release=false

if [ -z "$for_root" ]; then
    for_root=true
fi
if [ -z "$for_boot" ]; then
    for_boot=false
fi

# echo "get_release_body args: '$1' '$2' '$3' '$4' .." >&2

option=true
while [ $# -gt 0 ] && $option; do
    arg=$1
    case $arg in
        -v | --verbose)
            do_verbose=true;;
        -d | --delete)
            do_delete=true;;
        -s | --select)
            do_no_release=true;;
        -r | --root)
            for_root=true
            for_boot=false;;
        -b | --boot)
            for_root=false
            for_boot=true;;
        -f | --from)
            if [ $# -gt 0 ]; then
                release_from="$1"
                shift
            else
                do_help=true
            fi;;
        -h | -help | --help | \?)
            do_help=true;;
        *)
            option=false;;
    esac

    $option && shift
done

root_stage="$1"
boot_stage="$2"

if ! $for_root; then
    primary_stage="$boot_stage"
    primary_ops=bootops
    secondary_stage="$root_stage"
    secondary_ops=rootops
    for_secondary=$for_boot
else
    primary_stage="$root_stage"
    primary_ops=rootops
    secondary_stage="$boot_stage"
    secondary_ops=bootops
fi

if [ -z "$secondary_stage" ]; then
    for_secondary=false
else
    for_secondary=$for_root
fi

if $for_root && [ -z "$root_stage" ]; then
    echo "$cmd_name: no root staging directory given">&2
    do_help=true
fi
if $for_boot && [ -z "$boot_stage" ]; then
    echo "$cmd_name: no boot staging directory given">&2
    do_help=true
fi

filesudo=
[ "true" = "$do_filesudo" ] && filesudo=sudo

if $do_help; then
    help >&2
    exit 1
elif [ ! -d "$release_from" ]; then
    echo "$cmd_name: release directory '$release_from' does not exist">&2
    exit 2
elif [ ! -d "$release_from/files" ]; then
    echo "$cmd_name: no 'files' in release directory">&2
    exit 3
elif [ ! -d "$release_from/$primary_ops" ]; then
    echo "$cmd_name: no '$primary_ops' directory in release directory">&2
    exit 4
elif [ ! -x "$release_from/$primary_ops/filestore" ]; then
    echo "$cmd_name: no executable '$primary_ops/filestore' in release directory">&2
    exit 5
elif $for_secondary && [ -d "$release_from/$secondary_ops" -a ! -x "$release_from/$secondary_ops/filestore" ]; then
    echo "$cmd_name: no executable '$secondary_ops/filestore' in release directory">&2
    exit 6
elif [ ! -d "$primary_stage" ]; then
    echo "$cmd_name: directory to write release to does not exist - '$primary_stage'">&2
    exit 7
elif $for_secondary && [ -d "$release_from/$secondary_ops" -a ! -d "$secondary_stage" ]; then
    echo "$cmd_name: directory to write secondary release to does not exist - '$secondary_stage'">&2
    exit 8
elif $do_delete; then
    if $do_filecheck && [ ! -e "$primary_stage/files" ]; then
        echo "$cmd_name: no release in the stageing directory to delete">&2
        exit 9
    else
        $do_verbose && echo "$filesudo rm -rf \"$primary_stage/files\""
        if sudo $filesudo rm -rf "$primary_stage/files"; then
            echo "$cmd_name: release in staging directory deleted"
            if [ -d "$primary_stage/debug/" ]; then
                $do_verbose && echo "rm -rf \"$primary_stage/debug/\""
                rm -rf "$primary_stage/debug/"
            fi
            if [ -f "$primary_stage/filestore" ]; then
                $do_verbose && echo "rm -f \"$primary_stage/filestore\""
                rm -f "$primary_stage/filestore"
            fi
            exit 0
        else
            exit $?
        fi
    fi
else
    if $do_filecheck && ! $do_no_release && [ -e "$primary_stage/files" ]; then
        echo "$cmd_name: already a release in $primary_stage/files">&2
        echo "$cmd_name: destroy with -r <rel> -d first?">&2
        echo "$cmd_name: or create new ARM system - charmsys -n .. + chstage -p?">&2
        exit 10
    else
        if $do_no_release || make_release "$release_from" "$primary_stage"; then
            if ! $do_no_release && [ -d "$release_from/debug/" ]; then
                $do_verbose && echo "cp -rT \"$release_from/debug/\" \"$primary_stage/debug/\""
                cp -rT "$release_from/debug/" "$primary_stage/debug/"
            fi
            $do_verbose && echo "cp -rT --remove-destination \"$release_from/$primary_ops/\" \"$primary_stage/\""
            cp -rT --remove-destination "$release_from/$primary_ops/" "$primary_stage/"
            if $for_secondary; then
                if [ -d "$release_from/$secondary_ops" ]; then
                    $do_verbose && echo "ln -s \"$primary_stage/files\" \"$seconddary_stage/files\""
                    if ln -s "$primary_stage/files" "$secondary_stage/files"; then
                        $do_verbose && echo "cp -rT --remove-destination \"$release_from/$secondary_ops/\" \"$secondary_stage/\""
                        cp -rT --remove-destination "$release_from/$secondary_ops/" "$secondary_stage/"
                    fi
                else
                    if [ ! -e "$secondary_stage/filestore" ]; then
                        $do_verbose && echo "mkdir -p \"$secondary_stage/files\""
                        mkdir -p "$secondary_stage/files"
                        $do_verbose && echo "ln -s \"/home/suse/scripts/filestore_files.here\" \"$secondary_stage/filestore\""
                        ln -s "/home/suse/scripts/filestore_files.here" "$secondary_stage/filestore"
                    fi
                fi
            fi
        else
            rc=$?
            echo "$cmd_name: failed to copy whole release - rc $rc">&2
            exit $rc
        fi
        exit $?
    fi
fi
