#!/bin/bash

# functions to access permanent value sets from a bourne-like shell
# This file should be "sourced"

# Gray Girling, 2004

obj_trace=10
#obj_trace=1

. penv_env.sh

trace()
{   lev=$1
    prog="$2"
    shift
    if [ $obj_trace -le $lev ]; then
        echo "# ${prog} - $@" >&2
    fi
}

trace 1 $FUNCNAME "# OBJROOT set to '$OBJROOT'"

error()
{   echo "`basename "$0"`: $@" >&2
}

fail()
{   error "$@"
    exit 1
}

obj_set()
{
  _obj="$1"
  _var="_VL_${_obj}_$2"
  shift
  shift
  _val="$@"
  eval ${_var}=\"\$_val\"
}

obj_shset()
{
  _obj=$1
  _var=_VL_${_obj}_$2
  shift
  shift
  _val="$@"
  echo ${_var}=${_val}
}

obj_unset()
{
  _obj=$1
  _var=_VL_${_obj}_$2
  unset $_var
}

obj_store()
{ # echo "Obj $1 - set store for '$2' to dir '$3' type '$4'"
  obj_set "$1" "_${2}_vars" "$3"
  if [ ! -z "$4" ]; then
      obj_set "$1" "_${2}_type" "$4"
  fi
}


obj_val()
{
  _obj=$1
  _var=_VL_${_obj}_$2
  eval echo \$${_var}
}

obj_local()
{
  _obj=$1
  _var=_VL_${_obj}_$2
  eval ${3:-$2}=\$${_var}
}

obj_update()
{ 
  eval obj_set $1 $2 \$${3:-$2}
}

var_list()
{
  set | cut -b1-100 | awk -F= "/^$1/ { print substr(\$1, 1+length(\"$1\")); }"
}

obj_names()
{
  var_list "_VL_$1_$2"
}

obj_delete()
{
    obj=$1
    shift
    if [ $# -eq 0 ]; then
        names=`obj_names "$obj"`
    else
        names="$@"
    fi
    for name in `echo $names`; do
       # error "$FUNCNAME: unset _VL_${obj}_$name"
       unset _VL_${obj}_$name
    done
}

obj_marshal()
{
    obj="$1"
    shift
    if [ $# -eq 0 ]; then
        names=`obj_names "$obj"`
    else
        names="$@"
    fi
    for name in `echo $names`; do
       echo "$name `obj_val $obj $name`"
    done
}

obj_unmarshal()
{
    obj=$1
    rc=0
    while [ $rc -eq 0 ]; do
         read -r var val
         rc=$?
         if [ $rc -eq 0 -a "$var" != "" ]; then
             obj_set $obj $var $val
         # else
             # echo "Read failed: $rc"
         fi
    done
}

obj_show()
{   obj_marshal "$1"
}

obj_file()
{   eval echo \${_VLFILE_$1}
}

# warning: previous save file will continue to be used unless obj_unload used
obj_load()
{   local obj="$1"
    local file="$2"
    local rc=1
    
    trace 1 $FUNCNAME "loading obj $obj from '$file'"
    if [ -f "$file" ]; then
        obj_unmarshal "$obj" < "$file"
        if [ "" = "`obj_file $obj`" ]; then
            eval _VLFILE_${obj}=\"$file\"
            rc=0
        fi
    fi
    return $rc
}

obj_save()
{   local obj=$1
    local file

    if [ $# -gt 1 ]; then
        file="$2"
    else
        eval file=\${_VLFILE_$obj}
    fi
    if [ -z "$file" ]; then
        error "$obj - no file given to save object to"
	return 1
    elif [ ! -w "`dirname "$file"`" ]; then
        _dname="`dirname "$file"`"
	_dname=`cd "$_dname">/dev/null; pwd`
        error "$obj - `whoami` has no permission to write in $_dname"
	return 1
    else
        local tmp="`dirname "$file"`/penv_$obj_$$.env"
        # trace 1 $FUNCNAME "saving $obj via tmp file $tmp to $file"
	# try to do this as atomically as possible
        obj_marshal "$obj" > "$tmp"
	mv -f "$tmp" "$file"
        if [ "" = "`obj_file $obj`" ]; then
            trace 1 $FUNCNAME "resaved obj '$obj' to '$file'"
            eval _VLFILE_${obj}=\"$file\"
        else
            trace 1 $FUNCNAME "saved obj '$obj' to '$file'"
        fi
	return 0
    fi
}

obj_unload()
{   local obj="$1"
    obj_delete "$obj"
    unset _VLFILE_${obj}
}




# Permanent Environment object types


file_create()
{   # syntax: <file name>
    local dir="`dirname -- "$1"`"
    if mkdir -p "$dir"; then
        if [ ! -f "$1" ]; then
            touch "$1"
            trace 1 $FUNCNAME "created file '$1'"
            # ls -l $1
            return 0
        else
            return 1
        fi
    else
        error "can't create directory '$dir'" 
        return 1
    fi
}


type_open()
{   # syntax: <obj> <typename>
    local tobj="$1"
    local tname="$2"
    local typedir="`dirname "${OBJROOT}"`/../type"
    
    obj_load $tobj "${typedir}/${tname}.env"
    return $?
}


type_create()
{   local tobj="$1"
    local tname="$2"
    local tdescr="$3"
    local typedir="`dirname "${OBJROOT}"`/../type"
    local rc

    if file_create "${typedir}/${tname}.env"; then
        obj_load $tobj "${typedir}/${tname}.env"
        obj_set $tobj name "${tname}"
        obj_set $tobj _type "type"
        if [ ! -z "$tdescr" ]; then
            obj_set $tobj description "${tdescr}"
        fi
        return 0
    else
        return 1
    fi
}

type_field()
{   # syntax: <type obj> <field name> <default field value>
    local tobj="$1"
    local fname="$2"
    local fdefault="$3"
    obj_set $tobj "_field_$fname" "$fdefault"
}

type_instance()
{   # syntax: <type obj> <new obj>
    local tobj="$1"
    local nobj="$2"
    local fname
    
    shift
    shift
    if [ $# -eq 0 ]; then
        names=`type_fields "$tobj"`
    else
        names="$@"
    fi
    for fname in `echo $names`; do
        eval obj_set $nobj $fname `obj_val $tobj "_field_${fname}"`
    done
    eval obj_set $nobj _type `obj_val $tobj name`
}

type_store()
{   # syntax: <type obj> <field name> <default link> <default subtype>
    type_field "$1" "_${2}_vars" "$3"
    if [ ! -z "$4" ]; then
        type_field "$1" "_${2}_type" "$4"
    fi
}


type_fields()
{   obj_names "$1" "_field_"
}
            
    
type_close()
{   obj_save "$1"
    _rc=$?
    obj_unload "$1"
    return $_rc
}
    


# permanent environment files and environment names


firstdir()
{   ( IFS="/"
      eval set -- `echo $1`
      echo $1
    )
}

nofirstdir()
{   ( IFS="/"
      eval set -- `echo $1+`
      shift
      echo "$@" | tr ' ' '/' | tr -d '+' 
    )
}


objfile_create()
{   # syntax: <file name> <type>
    local fname="$1"
    local otype="$2"
    local dir="`dirname "$fname"`"
    
    if mkdir -p "$dir"; then
        if [ ! -f "$fname" ]; then
            touch "$fname"
            if [ ! -z "$otype" ]; then
                type_open _ttype "$otype"
                type_instance _ttype _nobj
                type_close _ttype
                if obj_save _nobj "$fname"; then
                    obj_unload _nobj
                    trace 2 $FUNCNAME "Obj: created file '$fname' of type '$otype'" 
                    return 0
                else
                    error "created file '$fname' of type '$otype'" 
                    return 1
                fi
            else
                trace 2 $FUNCNAME "Obj: created file '$fname'"
                return 1
            fi
        fi
    else
        error "can't create dir '$dir' for '$fname' of type '$otype'"
        return 1
    fi
}



objfile_getobjroot()
{   # syntax: objfile_getobjroot <objroot> <obj> [-c] <objname> [ <objtype> ]
    local objroot="$1"
    local obj=$2
    shift
    shift
    if [ "_$1" = "_-c" ]; then
        create=true
        shift
    else
        create=false
    fi
    local objname="$1"
    local objtype="$2"
    local origname="$objname"
    local objlink
    local objsel
    local dirobj
    local keeplooking=true

    trace 1 $FUNCNAME "Obj: $obj - open at root '$objroot' as '$objname'"
    while $keeplooking; do
        obj_unload $obj
        
        if [ ! -r "$objroot" -a "true" = "$create" ]; then
            objfile_create "$objroot" "$objtype"
        fi
        if [ -z "$objname" ]; then
            # got the object in question
            keeplooking=false
            trace 1 $FUNCNAME "Obj: $obj - found in file '$objroot'"
            echo "$objroot"
            return 0
        elif [ ! -r "$objroot" ]; then
            error "$obj - can't read environment file $objroot"
            keeplooking=false
        else
            obj_load "$obj" "$objroot"
            # trace 1 $FUNCNAME "Obj: $obj - loaded $objroot"
        
            trace 1 $FUNCNAME "Obj: $obj - looking up name '$objname' in $objroot"
            dirobj=`firstdir "$objname"`
            if [ "$dirobj" = "." ]; then
                trace 1 $FUNCNAME "Obj: $obj - dir '$dirobj' ignored"
                objname=`nofirstdir "$objname"`
            else
                objsel=`obj_val $obj ${dirobj} | tr ' /:,.' '_'`
                if [ -z "$objsel" ]; then
                    error "$obj - no value for ${dirobj} in ${objroot}"
                    obj_unload $obj
                    keeplooking=false
                else
                    objname=`nofirstdir "$objname"`
                    objlink=`obj_val $obj _${dirobj}_vars`
                    objtype=`obj_val $obj _${dirobj}_type`
                    if [ -z "$objlink" ]; then
                        objlink="$dirobj/"
                    fi
                    trace 1 $FUNCNAME "Obj: $obj - dirobj '$dirobj' objsel '$objsel' type '$objtype' link '$objlink'"
                    #echo link starts $objlink, root is $objroot
                    case "$objlink" in
                       /*)   objroot=${objlink}${objsel}.env;;
                       ../*) up=true
                             while $up; do
                                 objlink=`nofirstdir "$objlink"`
                                 objroot=`dirname "${objroot}"`
                                 case "$objlink" in
                                     ../*) ;;
                                     *) up=false;;
                                 esac
                                 trace 1 $FUNCNAME "Obj: $obj - in $objroot link '$objlink'"
                             done
                             #echo link ends $objlink, root is $objroot

                             objroot=`dirname "${objroot}"`/${objlink}${objsel}.env
                             #echo then root is $objroot
                             ;;
                       *)    objroot=`dirname "${objroot}"`/`basename "${objroot}" .env`/${objlink}${objsel}.env
                             #echo "normal exend '${objlink}${objsel}' makes root $objroot"
                            ;;
                    esac
                fi
            fi
            trace 1 $FUNCNAME "Obj: $obj - followed '$dirobj' to file $objroot - searching for '$objname'"
        fi
    done
    
    return 1 # failed
}


objfile_open()
{   # syntax: objfile_open <objroot> <obj> [-c] <objname> [ <objtype> ]
    
    local obj="$2"
    local objroot
    local rc
    
    if ! objroot=`objfile_getobjroot "$@"`; then
        rc=1
        error "$obj - can't open environment file for $3 $4 (rc $rc)"
    elif [ -z "$objroot" ]; then
        error "$obj - can't no environment file to open"
        rc=2
    else
        if [ ! -r "$objroot" ]; then
            error "$obj - can't read environment file $objroot"
            rc=3
        else
            obj_load "$obj" "$objroot"
            # trace 1 $FUNCNAME "Obj: $obj - loaded $objroot"
            rc=$?
        fi
    fi
    return $rc
}

    

objfile_file()
{   local root="$1"
    shift
    objfile_getobjroot "$root" _obj "$@"
    return $?
}

    

objfile_need()
{   # syntax: <objroot> <obj> <objname>
    local objroot="$1"
    local obj=$2
    local objname="$3"
    local rc=0
    if [ "" = "`obj_file $obj`" ]; then
        objfile_open "$objroot" "$obj" "$objname"
        rc=$?
    fi
    return $rc
}

objfile_set()
{   # syntax: <objroot> <obj> [-t <objtype> ] <objname> <value>...
    
    local objroot="$1"
    shift
    local obj=$1
    shift
    local objtype=
    if [ "_$1" = "_-t" ]; then
       shift
       objtype="$1"
       shift
    fi
    local objname="$1"
    shift

    trace 2 $FUNCNAME "Obj: $obj - root $objroot set '$objname' to '$@' type '$objtype'"
    objdir=`dirname "$objname"`

    objfile_open "$objroot" $obj -c "$objdir" "$objtype"
    if [ ! -z "`obj_file $obj`" ]; then
        objfield=`basename "$objname"`
        obj_set $obj "$objfield" "$@"
        if obj_save $obj; then
            return 0
	else
            error "failed to save new value for $objname"
	    return 1
	fi
    else
        error "failed to open $objname"
        return 1
    fi
}


objfile_store()
{
    local objroot="$1"
    local obj=$2
    local objname="$3"
    local objlink="$4"
    local objsubtype="$5"
    local objnewtype="$6"

    objdir=`dirname "$objname"`
    trace 2 $FUNCNAME "Obj: $obj - root $objroot set '$objname' link to '$objlink' (dir $objdir)"
    if ! objfile_open "$objroot" $obj -c "$objdir" "$objnewtype"; then
        return 2
    elif [ ! -z "`obj_file $obj`" ]; then
        objfield="`basename "$objname"`"
        obj_store "$obj" "$objfield" "$objlink" "$objsubtype"
        if obj_save "$obj"; then
            return 0
	else
            error "failed to save $objname"
	    return 1
	fi
    else
        error "failed to open $objname"
        return 1
    fi
}


objfile_unset()
{
    local objroot="$1"
    local obj=$2
    local objname="$3"

    trace 2 $FUNCNAME "Obj: $obj - root $objroot unset '$objname'"
    objdir=`dirname "$objname"`
    if ! objfile_open "$objroot" $obj "$objdir"; then
        return 2
    elif [ ! -z "`obj_file $obj`" ]; then
        objfield=`basename "$objname"`
        obj_unset $obj "$objfield"
        if obj_save $obj; then
            return 0
	else
            error "failed to unset $objname"
	    return 1
	fi
    else
        error "failed to open $objname"
        return 1
    fi
}


objfile_get()
{
    local objroot="$1"
    local obj=$2
    local objname="$3"

    trace 2 $FUNCNAME "Obj: $obj - root $objroot get '$objname'"
    objdir=`dirname "$objname"`
    if ! objfile_open "$objroot" $obj "$objdir"; then
        return 2
    elif [ ! -z "`obj_file $obj`" ]; then
        objfield=`basename "$objname"`
        obj_val $obj "$objfield" 
        return 0
    else
        error "$obj - failed to load $objname" 
        return 1
    fi
}


# Permanent Environment functions


penv_help()
{   cat <<EOF
penv commands:

    ls       [-a] <objname:dir>
             show content of name-value pairs in named object dir
    file     <objname:dir>
             show file name that should contain name-value pairs
    
    set      [-t <type>] <objname> <value>...
             assign value to name
    unset    <objname>
             remove value associated with name
    get      <objname>
             show value associated with name
    store    <objname:dir> <defaultdir> [<default subtype> [<newtype>]]
             associate name-value pair directory with object name

    objset   <object> [-t <type>] <name> <value> ...
             (as set but providing local env object used to cache details)
    objunset <object> <objname>
             (as unset but providing local env object used to cache details)
    objget   <object> <objname>
             (as get but providing local env object used to cache details)
    objstore <object> <objname> <defaultdir> [<default subtype> [<newtype>]]
             (as store but providing local env object used to cache details)
        
    create   <object> <objname> [<type>]
    open     <object> <objname>
    needopen <object> <objname>
    close    <object>
EOF
}

penv_create()
{   # syntax: <object> <objname> [<type>]
    local obj="$1"
    shift
    objfile_open "$OBJROOT" "$obj" -c "$@"
    return $?
}

penv_open()
{   # syntax: penv_open <obj> [-c] <objname> [ <objtype> ]
    objfile_open "$OBJROOT" "$@"
    return $?
}

penv_needopen()
{   # syntax: <object> <objname>
    objfile_need "$OBJROOT" "$@"
}

penv_close()
{   # syntax: <object>
    obj_save "$1"
    _rc=$?
    obj_unload "$1"
    return $_rc
}

penv_ls()
{   # syntax: [-a] <object "directory" name>
    local all=false
    if [ "_$1" = "_-a" ]; then
        all=true
        shift
    fi
    if penv_open _dir "$1"; then
        if $all; then
            obj_show _dir
        else
            obj_show _dir | grep -v "^_"
        fi
        obj_unload _dir
        return $?
    else
        return $?
    fi
}

penv_file()
{   # $1 = object "directory" name
    objfile_file "$OBJROOT" "$1" 
}

penv_dir()
{   # $1 = object "directory" name
    local file=`objfile_file "$OBJROOT" "$1"`
    if [ -z "$file" ]; then
        return 1
    else
        dirname "$file"
    fi
}

penv_all()
{   # $1 = object "directory" name
    local file=`objfile_file "$OBJROOT" "$1"`
    if [ -z "$file" ]; then
        return 1
    else
        ls `dirname "$file"` | grep ".env$"
        # ls | awk "/.env$/ { sub(/.env$/, \"\"); print; }" 
    fi
}

penv_objset()
{   # syntax: <obj> [-t <type>] <name> <value>...
    if [ $# -lt 2 ]; then
        error "syntax: [-t <type>] <name> <val>..."
        return 1
    else
        objfile_set "$OBJROOT" "$@"
        return 0
    fi
}

penv_set()
{   # syntax: [-t <type>] <name> <value>...
    if penv_objset _obj "$@"; then
        obj_unload _obj
    fi
}

penv_objunset()
{   # $1 = name, $2... value
    local obj="$1"
    local objname="$2"
    if [ "" = "$objname" ]; then
        error "syntax: <name>"
        return 1
    else
        objfile_unset "$OBJROOT" "$obj" "$objname"
        return 0
    fi
}

penv_unset()
{   # $1 = name
    if penv_objunset _obj "$1"; then
        obj_unload _obj
	trace 3 $FUNCNAME "Obj: unset $1"
    fi
}

penv_objget()
{   # $1 = object,$2 = name
    local obj="$1"
    local objname="$2"
    if [ "" = "$objname" ]; then
        error "syntax: <obj> <name>"
        return 1
    else
        objfile_get "$OBJROOT" "$obj" "$objname"
        return $?
    fi
}


penv_get()
{   # $1 = name
    if penv_objget _obj "$1"; then
        obj_unload _obj
	trace 3 $FUNCNAME "Obj: got $1"
        return 0
    else
        return 1
    fi
}

penv_objstore()
{   # syntax: <obj> <name> <dir> [<subtype> [<newtype>]]
    
    if [ $# -lt 3 ]; then
        error "syntax: <obj> <name> <dir> [<subtype> [<newtype>]] "
        return 1
    else
        objfile_store "$OBJROOT" "$@"
        return 0
    fi
}

penv_store()
{   # Syntax: <objname> <default directory> [<default subtype> [<newtype>]]
    if penv_objstore _obj "$@"; then
        obj_unload _obj
    fi
}


adhoc_update_mcs()
{   
    obj_trace=3
    for f in `ls -d $DIR_PENV/machine/*.env`; do
	echo
	echo "$f:"
        if obj_load   _tmp "$f"; then
            # obj_store  _tmp kernel ../kernel/
            # obj_set    _tmp flavour debug

            obj_save   _tmp
            obj_show   _tmp
            obj_unload _tmp
        fi
    done
    obj_trace=3

}
